/*
    Copyright (C) 2004 Mark Boyd

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be fun to play,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include<stdlib.h>
#include<string.h>
#include<stdio.h>
#include<math.h>
#include<time.h>
#include<iostream>
#include<allegro.h>
#include<deque>
#include<list>
#include<string>
#include<sstream>

using std::string;
using std::min;
using std::max;

#include"uiobject.h"
#include"tanx.h"
#include"tanxdat.h"


typedef unsigned char byte;


//#define SPEEEED

//Possible terrain types.  There are many different kinds of "dirt", but they
//all behave in the same way.
enum
{ 
  T_EMPTY_SPACE = 0,
  T_DIRT_MIN = 1,
  T_DIRT_MAX = 63,
  T_SHOT = 64,
  T_WALL = 15, //walls are made of dirt.
  T_RUBBER = 69,
  T_GREEN_STUFF = 66,
  T_TANK_COLOUR = 68,
};

bool is_dirt(int x) {return x>=T_DIRT_MIN && x<=T_DIRT_MAX;}
bool can_fall(int x) {return is_dirt(x);}

std::string int2str(int i)
{
  std::stringstream ss;
  ss<<i;
  return ss.str();
}

std::string float2str(double d, int chars=5)
{
  std::stringstream ss;
  ss<<d;
  return ss.str().substr(0, chars);
}

//shot types
enum {S_STANDARD,S_AIRBURST,S_PONGER,S_SINKER,S_FLOATER,S_GROUNDBURST};
const int non_special_shots(3);

//warheads
enum {W_TINY,W_SMALL,W_MEDIUM,W_LARGE,W_HUGE,W_DIRTBUSTER,W_HOLE,W_SINKWALL,W_WALL,W_CLOUD,W_DIRTBALL,W_RUBBER};
const int non_special_warheads(7);


char *shots[] = {"Standard","Airburst","Ponger","Sinker","Floater","Groundburst"};
char *warheads[] = {"Tiny","Small","Medium","Large","Huge","Dirtbuster",
                    "Hole","Sinkwall","Wall","Cloud","Dirtball","Rubber"};

//Limit use of some warhead and shot types to prevent a defensive stalemate situation.
bool is_special_warhead (int i)
{
  switch(i)
  {
    case W_WALL:
    case W_CLOUD:
    case W_DIRTBALL:
    case W_RUBBER:

    return true;
  }

  return false;
}

bool is_special_shot (int i)
{
  switch(i)
  {
    case S_SINKER:
    case S_FLOATER:
    case S_GROUNDBURST:

    return true;
  }

  return false;
}

//returns a random number in the range 0<=x<i
//intended to be a better version of rand()%i
int random(int i)
{
  return int(double(rand())*i/RAND_MAX);
}

DATAFILE *sound::df(0);
int sound::ok(0);
int sound::d;
int sound::m;

int sound::init()
{
  if(df){ok=1;return 0;}//Already initialised
  ok=!install_sound(DIGI_AUTODETECT,MIDI_NONE,0);
  if(!ok)return 1;
  reserve_voices(2,0);
  ok=!!(df=load_datafile("#"));
  if(!ok)return 1;

  return 0;
}

void sound::init_off(){ok=0;}

void sound::finish()
{
  if(df)
    unload_datafile(df);
}

int sound::play(int i,int vol,int pan,int freq,int loop)
{
  return ok?play_sample((SAMPLE *)(df[i].dat),vol,pan,freq,loop):0;
}

void sound::stop(int i)
{
  stop_sample((SAMPLE *)(df[i].dat));
}

//The ground
//m_width, m_depth = total width/depth
//m_left,m_top,m_right,m_bottom = left, top etc of the part which shows on the screen

ground::ground(int w,int d,int l,int t,int r,int bm)
:m_width(w),m_depth(d),m_left(l),m_top(t),m_right(r),m_bottom(bm)
{
  const int dirt_mid = (T_DIRT_MIN+T_DIRT_MAX)/2;
  const double dirt_range = (T_DIRT_MAX-T_DIRT_MIN)/3;

  m_xdisp=m_ydisp=0;
  m_bitmap=create_bitmap(m_width+1,m_depth+1);
  m_buffer=create_bitmap(m_right-m_left+1,m_bottom-m_top+1);

  m_sky_table=new byte[2*m_depth+1];
  for(int a=0;a<=2*m_depth;a++)
    m_sky_table[a]=255-(byte)(63.0*exp(-3.0*a/m_depth));
 
  clear_to_color(m_bitmap, T_EMPTY_SPACE);
  int height=random((3*m_depth/5)+m_depth/5);
  for(int x=0;x<=m_width;x++)
  {
    for(int h=height;h<=m_depth;h++)
      m_bitmap->line[h][x]=static_cast<byte>(
       dirt_mid+dirt_range*cos(((double)(h+random(3)+random(3)))/25.0));
    height+=(random(5)-2);
    height=max(m_depth/5,min(height,4*m_depth/5));
  }
  rect(screen,m_left-1,m_top-1,m_right+1,m_bottom+1,64);
  draw();
}

ground::~ground()
{
  delete[] m_sky_table;
  destroy_bitmap(m_bitmap);
  destroy_bitmap(m_buffer);
}

void ground::draw()
{
  //Draw sky onto buffer.  Sky moves half the distance that the ground moves
  int view_height=m_bottom-m_top+1;
  int view_width=m_right-m_left+1;
  for(int y=0;y<view_height;y++)
    hline(m_buffer,0,y,view_width,m_sky_table[m_ydisp/2+y]);

  //draw ground on top.
  masked_blit(m_bitmap,m_buffer,m_xdisp,m_ydisp,0,0,view_width,view_height);

  //blit it all to the screen
#ifdef DOS_FULLSCREEN
  vsync();
#else
  rest(10);
#endif
  blit(m_buffer,screen,0,0,m_left,m_top,view_width,view_height);
}

//instantly moves the view to a location
void ground::moveto(int x,int y)
{
  x=x-(m_right-m_left)/2;
  y=y-(m_bottom-m_top)/2;
  m_xdisp=max(0,min(x,m_width-m_right+m_left));
  m_ydisp=max(0,min(y,m_depth-m_bottom+m_top));
}

//Fancy scroll.  Scroll 1 pixel on the first frame, then 2 on the next, then 3,
//but slow down after half way so we scroll 1(ish) pixel on the last frame too.
void ground::scrollto(int x,int y)
{
  x=x-(m_right-m_left)/2;
  y=y-(m_bottom-m_top)/2;
  int dx=max(0,min(x,m_width-m_right+m_left))-m_xdisp;
  int dy=max(0,min(y,m_depth-m_bottom+m_top))-m_ydisp;
  int xold=m_xdisp;
  int yold=m_ydisp;
  int distance=max(abs(dx),abs(dy));
  int m=1;
  for(int n=1;n<distance;n=min(n+m,distance))
  {
    m_xdisp=(int)(xold+n*dx/distance);
    m_ydisp=(int)(yold+n*dy/distance);
    draw();
    m=max(m+((n<distance/2)?1:-1), 1);
  }
  m_xdisp=xold+dx;
  m_ydisp=yold+dy;
  draw();
}


void ground::addplayer(int n,int nplayers,int *xp,int *yp)
{
  *xp=(int)((m_width*(2*nplayers-1))/(n*2));
  int h=0;
  while(!(m_bitmap->line[h][(*xp)-1]||m_bitmap->line[h][*xp]||
          m_bitmap->line[h][(*xp)+1]))h++;
  *yp=h-3;
}

void ground::drawplayer(player *pp)
{
  int x,y;
  pp->getcoords(&x,&y);
  for(int a=x-1;a<=x+1;a++)
    for(int b=y+1;b<=y+2;b++)
      m_bitmap->line[b][a]=T_TANK_COLOUR;
}

//replace a (dead) tank with dirt.
void ground::drawoff(player *pp)
{
  int x,y;
  pp->getcoords(&x,&y);
  for(int a=x-1;a<=x+1;a++)
    for(int b=y+1;b<=y+2;b++)
      m_bitmap->line[b][a]=T_DIRT_MAX;//dark dirt
}

//Make the player fall one pixel, if there is empty space to fall into
//return true if the player falls, false if not
bool ground::fallplayer(player *pp)
{
  int a,x,y;
  pp->getcoords(&x,&y);
  bool can_fall=(y+3<=m_depth)&&(!m_bitmap->line[y+3][x-1]&&!m_bitmap->line[y+3][x]&&
                                 !m_bitmap->line[y+3][x+1]);
  if(can_fall)
  {
    pp->down1();y++;
    for(a=x-1;a<=x+1;a++)
    {
      m_bitmap->line[y+2][a]=68;
      m_bitmap->line[y][a]=0;
    }
  }
  return can_fall;
}

//fire a shot.  Returns true if the shot expoldes.
//If is_real is false, nothing actually happens. 
//if x_bounce, y_bounce are not nul, they recieve te coords of the first bounce
bool ground::fire(const int x_start, const int y_start, double x_speed ,double y_speed,
                  int shot, int *x_end, int *y_end, const int is_real, int *x_bounce, int *y_bounce)
{
  int time,count,boings(0),pongs(0);
  
  const double g=0.1;  //gravity
  const double speed_factor = 0.01;
  time=0;
  int x=x_start, y=y_start;
  double dox=x_start, doy=y_start;
  count=0;

  bool tunneller=(shot==S_SINKER||shot==S_FLOATER||shot==S_GROUNDBURST);
  bool floater = (shot==S_FLOATER||shot==S_GROUNDBURST);
  bool ponger=(shot==S_PONGER);
  bool bang=false;
  bool bounced=false;
  

  //main loop - make shot move, interact with ground
  while(x>=0&&x<=m_width&&y<=m_depth&&y>=-1500&&!bang&&boings<10)
  {  
    byte b = (y>=0)?m_bitmap->line[y][x]:0;
    
    if(b==T_EMPTY_SPACE || (is_dirt(b)&&tunneller) || b==T_RUBBER)
    {
      //did not explode
      byte new_terrain = b;

      if(y>=0)
      {
        if(b==T_RUBBER)
        {
          //reverse direction (so shot has a fair chance of hitting the sender)
          y_speed=ponger?-y_speed:g-y_speed;
          x_speed=-x_speed;
          new_terrain = T_GREEN_STUFF; //rubber can only be used once, this avoids infinite loops.
          if(is_real)sound::play(BOING,255,128,1000,0);else boings++;
          if (!bounced)
          {
            bounced = true;
            if (x_bounce) *x_bounce=x;
            if (y_bounce) *y_bounce=y;
          }
        }        
        if(is_real) m_bitmap->line[y][x]=T_SHOT;
      }
      if(is_real)
      {
        moveto(x,y);
        if(!(++count%5))draw();
        if(tunneller && is_dirt(b))
          new_terrain = (T_DIRT_MIN+T_DIRT_MAX)/2; //tunnellers alter terrain
        if(y>=0) m_bitmap->line[y][x]=new_terrain;
      }
      time++;

      x = int(dox=dox+x_speed*speed_factor);
      y = int(doy=doy+y_speed*speed_factor);

      //apply gravity
      if(!ponger)
      {
        if(is_dirt(new_terrain)&&floater )y_speed-=g; else y_speed+=g;
      }

      if(y_speed>0 && shot==S_AIRBURST) bang=true;
      if(y_speed<0 && shot==S_GROUNDBURST && is_dirt(b)) bang=true;
    }
    else bang=true;

    //Special bouncing code for when pongers hit the edges
    if(ponger)
    {
      if(x<0)
      {
        x=int(dox=max(0.0,-dox));
        x_speed=-x_speed;
        if(is_real)sound::play(PONG2,255,128,1000,0);
        pongs++;
      }
      if(x>m_width)
      {
        x=int(dox=min<double>(m_width,2*m_width-dox));
        x_speed=-x_speed;
        if(is_real)sound::play(PONG2,255,128,1000,0);
        pongs++;
      }
      if(y<0)
      {
        y=int(doy=max(0.0,-doy));
        y_speed=-y_speed;
        if(is_real)
        sound::play(PONG2,255,128,1000,0);
        pongs++;
      }
      if(pongs>15)
      {
        //Punish timewasters.  This should cause the shot to bounce off
        //the ceiling and hit the tank which fired it.
        x_speed=double(x_start-dox)/4.0;
        y_speed=-double(1+y_start+doy)/4.0;
        pongs=0; 
      }
    }
  }

  if(is_real)draw();

  //deal with shots below the bottom...
  if(x>=0&&x<=m_width&&y>m_depth)
  {
    y=m_depth+1;//One lower than the bottom row of the bitmap
    bang=true;
  }

  if(!bang)
  {
    //we're out of the loop but didn't explode.  Must have gone off the sides or off the top.
    if(is_real)
    {
      if(y>-1500)
      {
        //Off the sides.  This sound is intended to be a car alarm.  
        rest(1500);
        sound::play(BANG,255,128,1000,0);
        rest(200);
        sound::play(SQUEAL,255,128,1000,1);
        rest(1500);
        sound::stop(SQUEAL);
      }

      boringbox bb(100,100,220,135,64,y>-1500?"Miles Away!":"Into Orbit!");
      bb.spawn(new button(190,120,210,130,64,"OK",stoppit1));
      bb.redraw();
      uiobject::uol.wait();
    }
  }

  //"return" final position
  *x_end=x;
  *y_end=y;

  if (!bounced)
  {
    if (x_bounce) *x_bounce=x;
    if (y_bounce) *y_bounce=y;
  }

  return bang;
}

void ground::detonate(players *pp,int x,int y,int warhead)
{
  int size=0;
  switch(warhead)
  {
  case W_HUGE:   size++;
  case W_LARGE:  size++;
  case W_MEDIUM: size++;
  case W_SMALL:  size++;
  case W_TINY:   size++;
    circlebang(x,y,8*(size),(6-size)*20,*pp);
    break;

  case W_DIRTBUSTER:
    sound::play(BUST,255,128,1000,0);
    for(int yy=min(y,m_depth);yy>max(y-45,0);yy--)
    {
      for(int xx=max(0,x+yy-y-2);xx<=min(m_width,x+y-yy+2);xx++)
        m_bitmap->line[yy][xx]=64;
      rest(4);
      pp->redrawall();
      draw();
    }
    for(int yy=min(y,m_depth);yy>max(y-45,0);yy--)
    {
      for(int xx=max(0,x+yy-y-2);xx<=min(m_width,x+y-yy+2);xx++)
        m_bitmap->line[yy][xx]=0;
      rest(4);
      pp->redrawall();
      draw();
    }
    landfall(x-46,x+46,pp);
    rest(500);
    break;
  
  case W_HOLE:
    if(y-8>=0)
    for(int xx=max(0,x-3);xx<=min(m_width,x+3);xx++)
      if(m_bitmap->line[y-8][xx])m_bitmap->line[y-8][xx]=T_GREEN_STUFF;
    if(y-7>=0)
    {
      if(x-5>=0       && m_bitmap->line[y-7][x-5]) m_bitmap->line[y-7][x-5]=T_GREEN_STUFF;
      if(x-4>=0       && m_bitmap->line[y-7][x-4]) m_bitmap->line[y-7][x-4]=T_GREEN_STUFF;
      if(x+4<=m_width && m_bitmap->line[y-7][x+4]) m_bitmap->line[y-7][x+4]=T_GREEN_STUFF;
      if(x-5<=m_width && m_bitmap->line[y-7][x+5]) m_bitmap->line[y-7][x+5]=T_GREEN_STUFF;
    }
    circlebang(x,y,8,5,*pp);
    break;

  case W_WALL:
    sound::play(WALL,255,128,1000,1);
    for(int h=y-1;h>=0;/*donothing*/)
    {
      m_bitmap->line[h][x]=WALL;
      rest(4);
      moveto(x,h);
      if(h--)m_bitmap->line[h][x]=64;
      draw();
    }
    sound::stop(WALL);
    pp->redrawall();
    landfall(x-2,x+2,pp);
    rest(500);
    break;

  case W_SINKWALL:
    {
      byte old_terrain=0;
      sound::play(WALL,255,128,1000,1);
      for(int h=y-1;h<=m_depth;/*donothing*/)
      {
        if(is_dirt(old_terrain))
          m_bitmap->line[h][x]=T_GREEN_STUFF;
        else
          m_bitmap->line[h][x]=old_terrain;
        rest(4);
        moveto(x,h);  
        if(++h<=m_depth)
        {
           old_terrain = m_bitmap->line[h][x];
           m_bitmap->line[h][x]=64;
        }
        draw();
      }
      sound::stop(WALL);
      pp->redrawall();
      rest(500);
    }
    break;

  
  case W_CLOUD:
    {
      const int radius=40;
      sound::play(SPLASH,255,128,1000,0);
      for(int xx=max(0,x-radius);xx<=min(m_width,x+radius);xx++)
        for(int yy=max(0,(int)(y-::sqrt(radius*radius-(x-xx)*(x-xx))));
            yy<=min(m_depth,(int)(y+sqrt(radius*radius-(x-xx)*(x-xx)))); yy++)
        if(!(random(15)))
          m_bitmap->line[yy][xx]=T_GREEN_STUFF;
      pp->redrawall();
      draw();
      rest(1000);
    }
    break;

    case W_DIRTBALL:
    {   
      const int radius=35;
      sound::play(BALL,255,128,1000,0);
      for(int a=0;a<radius;a++)
      {
        for(int b=a;b>=0;b--)
        annulus(x,y,0,b,T_DIRT_MIN+int((T_DIRT_MAX-T_DIRT_MIN)*(0.75-0.5*cos(M_PI*(b+radius-a)/(2*radius)))));
        pp->redrawall();
        rest(10);
        draw();
      }
      landfall(x-radius,x+radius,pp);
      rest(500);
    }
    break;

    case W_RUBBER:
    {   
      const int radius=40;
      const int inner_radius=35;
      sound::play(SPLASH,255,128,1000,0);
      for(int c=0;c<=radius;c++)
      {
        for(int xx=max(0,x-radius);xx<=min(m_width,x+radius);xx++)
          for(int yy=max(0,y-radius);yy<=min(m_depth,y+radius);yy++)
            if((xx-x)*(xx-x)+(yy-y)*(yy-y)<c*c&&m_bitmap->line[yy][xx]&&
              c+random(c)<inner_radius) m_bitmap->line[yy][xx]=T_RUBBER;
        pp->redrawall();
        draw();
      }
      rest(1000);
    }
    break;
  }
}

//a circular explosion.
void ground::circlebang(int x,int y,int radius,int bd,players &the_players)
{
  sound::play(BANG,min(6*radius,255),128,1200-10*radius,0);

  //detect players in blast radius
  
  players::iterator i=the_players.first_player();
  do
  {
    int player_x, player_y;
    (*i)->getcoords(&player_x,&player_y);
    for(int xx=player_x-1;xx<player_x+2;xx++)
      for(int yy=player_y+1;yy<player_y+3;yy++)
        if((xx-x)*(xx-x)+(yy-y)*(yy-y)<radius*radius)
          {
          (*i)->blast(bd);
          goto loop_end;
          }
    loop_end:
    (void)"label must be followed by statement";
  }while((i=the_players.next_player(i)) != the_players.first_player());

  //do the actual explosion
  for(int a=20;a>=0;a--)
  {
    for(int b=48;b>=0;b--)
      annulus(x,y,0,(int)((b)*(400-a*a)*radius/400/48),176-b);
    the_players.redrawall();
    draw();
  }
  for(int a=1;a<=20;a++)
  {
    annulus(x,y,(400-a*a)*radius/400,radius,0);
    the_players.redrawall();
    draw();
  }

  //apply gravity
  landfall(x-radius,x+radius,&the_players);
  rest(1000);
}

//make "hanging" bits of ground fall down (after a ground-destroying explosion)
//we maintain a list of contiguous vertical strips of pixels.
struct strip
{
  int top;
  int bottom;
  int x;
  int lives;
  strip(int t, int b, int x_):top(t),bottom(b),x(x_),lives(2){}
};

void ground::landfall(int xmin,int xmax,players *pp)
{
  std::list<strip> strips;
  xmin=max(0,xmin-2);
  xmax=min(m_width,xmax+2);
  for(int x=xmin;x<=xmax;x++)
  {
    int y=m_depth;
    while(y>=0)
    {
      while(y>=0&&can_fall(m_bitmap->line[y][x]))y--;
      while(y>=0&&!can_fall(m_bitmap->line[y][x]))y--;
      if(y>=0&&is_dirt(m_bitmap->line[y][x]))
      {
        int strip_bottom = y;
        while(y>=0&&can_fall(m_bitmap->line[y][x]))y--;
        strips.push_back(strip(y, strip_bottom, x));
      }
    }
  }

  int c=1;
  while(strips.size())
  {
    for(std::list<strip>::iterator i=strips.begin(); i!=strips.end(); )
    {
      if(i->bottom<m_depth && m_bitmap->line[i->bottom+1][i->x] == T_EMPTY_SPACE)
      {
        for(int y=i->bottom; y>i->top+1; --y)
          m_bitmap->line[y+1][i->x] = m_bitmap->line[y][i->x];
        m_bitmap->line[i->top][i->x]=0;
        i->top++;
        i->bottom++;
        ++i;
      }
      else
      {
        //"lives" is a bit of a hack: the strip may be sitting on top of a player
        //if that is the case, it won't fall on the first iteration but will on the second,
        //so make sure it doesn't get deleted too early.
        if(i->lives)
        {
          i->lives--;
          ++i;
        }
        else  
        {
          //postincrement is cool.
          strips.erase(i++);
        }
      }
    }
    pp->fall();
    rest(60/c++);
    draw();
  }


  //players may continue to fall after ground has finished falling...
  while(pp->fall())
  {
    rest(60/c++);
    draw();
  }
  pp->rbcheck(m_depth-2);
  draw();
}



void ground::annulus(int x,int y,int inner,int outer,int col)
{
  for(int a=max(0,y-outer+1);a<=min(m_depth,y+outer-1);a++)
  {
    int tmp=min(outer-1,(int)sqrt(outer*outer-(y-a)*(y-a)));
    int left_out=max(0,x-tmp);
    int right_out=min(m_width,x+tmp);
   
    if(inner*inner>(y-a)*(y-a))
    {
      int tmp=min(inner-1,(int)sqrt(inner*inner-(y-a)*(y-a)));
      int left_in=max(0,x-tmp);
      int right_in=min(m_width,x+tmp);

      if(left_in>left_out)   hline(m_bitmap,left_out,a,left_in,col);
      if(right_in<right_out) hline(m_bitmap,right_in,a,right_out,col);
    }
    else
      hline(m_bitmap,left_out,a,right_out,col);
  }
}

int ground::getdepth(void){return m_depth;}
int ground::getwidth(void){return m_width;}
byte ground::getpixel(int x,int y){return m_bitmap->line[y][x];}


//"construct" a player.  Get name and type (human or computer) from UI, then
//place the tank and draw it.
player::player(int n,int pn,ground *gp)
{
  //things we can initialise immediately
  m_armour=500;
  m_score=0;
  m_kills=0;
  m_has_killed_self=false;
  m_blast_damage=m_fall_damage=0;
  m_rock_bottom=false;
  m_nspecials=10;
  m_firing_data=0;

  strcpy(m_defaults.power,"60");
  strcpy(m_defaults.angle,"45");
  m_defaults.direction=1;
  m_defaults.shot=S_STANDARD;
  m_defaults.warhead=W_MEDIUM;
  m_defaults.left=40;
  m_defaults.top=0;

  //get name and type (human/computer)
  textbox *tb;
  menu *m;
  boringbox bb(70,50,270,150,64,"Who are you?");
  bb.spawn(new text(90,72,64,"Name:"));
  bb.spawn(new text(90,92,64,"Type:"));
  bb.spawn(tb=new textbox(170,70,10,64,65,(string("Player ")+int2str(pn)).c_str()));
  char *horc[2]={"Human","Computer"};
  bb.spawn(m=new menu(170,90,64,2,1,horc));
  bb.spawn(new button(240,130,260,140,64,"OK",stoppit1));
  bb.redraw();
  #ifndef SPEEEED
  uiobject::uol.wait();
  #endif
  m_name=tb->getstr();
  m_type = m->getcurrent()?COMPUTER:HUMAN;

  gp->addplayer(n,pn,&m_x,&m_y);
  gp->drawplayer(this);
  gp->draw();
}

player::~player()
{
  if(m_firing_data) delete m_firing_data;
}

//Get data for firing (power, angle, shot type...)
//Different procedure depending on whether the player is human or not.
void player::fire(double *dxp,double *dyp,int *shot,int *warhead, players *pp,
                  ground *gp,bool *stubborn)
{
  if(m_type == COMPUTER)
  {
    computer_fire(dxp, dyp,shot, warhead, pp, gp);
    *stubborn = false;
  }
  else
  {
    human_fire(dxp, dyp,shot, warhead, pp, gp, stubborn);
  }
  if(is_special_warhead(*warhead) || is_special_shot(*shot))
  {
    m_nspecials--;
    if(m_nspecials==0)
    {
      if(is_special_shot(*shot)) m_defaults.shot = S_STANDARD;
      if(is_special_warhead(*warhead)) m_defaults.warhead = W_MEDIUM;
    }
  }
}

//AI shot selection.  This uses a brainless brute-force search algorithm

void player::computer_fire(double *dxp,double *dyp,int *shot,int *warhead,
                           players *pp, ground *gp)
{
  if (!computer_use_old_firing_data(dxp, dyp, shot, warhead, pp, gp))
  {
    computer_choose_shot(dxp, dyp, shot, warhead, pp, gp);
  }

  //Shot is now chosen - display the choice
  double pow = sqrt(*dxp**dxp + *dyp**dyp);
  double ang = (*dxp||*dyp)?atan2(*dyp,std::abs(*dxp)):0;

  boringbox bb(m_defaults.left,m_defaults.top,m_defaults.left+240,
               m_defaults.top+110,64,("It's "+m_name+"'s turn.").c_str());

  bb.spawn(new text(m_defaults.left+10,m_defaults.top+20,64,
           (string("Score: ")+int2str(m_score)+" Armour: "+int2str(m_armour)).c_str()));

  bb.spawn(new text(m_defaults.left+10,m_defaults.top+30,64,
            (string("Specials: ")+int2str(m_nspecials)).c_str()));

  bb.spawn(new text(m_defaults.left+10,m_defaults.top+48,64,
             (string("Power: ")+float2str(pow)).c_str()));

  bb.spawn(new text(m_defaults.left+10,m_defaults.top+63,64,
           (string("Angle: ")+float2str(ang*180/M_PI)+" "+(*dxp>0?"Right":"Left")).c_str()));

  bb.spawn(new text(m_defaults.left+10,m_defaults.top+78,64, (string("Shot: ")+shots[*shot]).c_str()));
  bb.spawn(new text(m_defaults.left+10,m_defaults.top+93,64, (string("Warhead: ")+warheads[*warhead]).c_str()));

  bb.spawn(new button(m_defaults.left+170,m_defaults.top+60,m_defaults.left+220,
                      m_defaults.top+90,68,"FIRE!",stoppit1));
  bb.redraw();
  #ifndef SPEEEED
  uiobject::uol.wait();
  #else
  rest(700);
  #endif

  //store defaults so the box wil appear in the same place next time
  m_defaults.left=bb.getl();
  m_defaults.top=bb.gett();
}

void player::computer_choose_shot(double *dxp,double *dyp,int *shot,int *warhead,
                                  players *pp, ground *gp)
{
  int x_target,y_target;
  pp->getvictim(this,&x_target,&y_target);
  x_target-=m_x;
  y_target-=(m_y+1);

  *shot = random(m_nspecials?sizeof(shots)/sizeof(*shots):non_special_shots);
  *warhead = 0; //Since we're just trying trajectories, we don't care about this right now.
  if((gp->getpixel(m_x,m_y)&&*shot!=S_SINKER&&*shot!=S_FLOATER)||
     (gp->getpixel(m_x,m_y)>63&&gp->getpixel(m_x,m_y)))*warhead=W_DIRTBUSTER;

  //For most warheads, we want to explode under (or near) the target.  But the dirtball is
  //best exploded above the target.
  int smallest_error=10000000;
  int smallest_dball_error = 10000000;
  int error, dball_error;
  double pow=45, ang=M_PI/4;
  double best_pow=pow, best_ang = ang;
  double best_dball_pow=pow, best_dball_ang = ang;
  int trytp, tryxf, tryyf;

  size_t t1=time(0);
  while(time(0)-t1<2);
  {
    for(int a=0;a<100;a++)
    {
      doitagain:
      pow=random(90)+10;
      ang=double(M_PI*rand()/RAND_MAX/2);
      if(*shot==S_PONGER){ang*=2;pow=42;}
      *dxp=pow*cos(ang);
      if((*dxp)<15&&(*shot!=S_PONGER))goto doitagain;
      if(ang<0.01&&*shot==S_PONGER)goto doitagain;
      *dxp*=(x_target>0?1:-1);
      *dyp=pow*sin(ang);
      trytp=*shot;

      if(!gp->fire(m_x,m_y,*dxp,-*dyp,trytp,&tryxf,&tryyf,0))
        error=9999999;
      else
      {
        //tryxf, tryyf = where this simulated shot explode
        int too_low = max(0, tryyf-m_y-y_target);
        int too_high = max(0, m_y+y_target-tryyf);
        error=(tryxf-m_x-x_target)*(tryxf-m_x-x_target)+too_high*too_high;
        dball_error =(tryxf-m_x-x_target)*(tryxf-m_x-x_target)+too_low*too_low;
      }
      if(error<smallest_error)
      {
        smallest_error=error;
        best_pow=pow;
        best_ang=ang;
      }
      if(dball_error<smallest_dball_error && tryyf>=0)
      {
        smallest_dball_error=dball_error;
        best_dball_pow=pow;
        best_dball_ang=ang;
      }
    }
  }

  //choose an offensive weapon
  do
  {
    *warhead = random(sizeof(warheads)/sizeof(*warheads));
  } while (*warhead == W_DIRTBUSTER || *warhead == W_HOLE || *warhead == W_RUBBER || *warhead==W_SINKWALL);
  if(*warhead == W_DIRTBALL)
  {
    smallest_error = smallest_dball_error;
    best_pow = best_dball_pow;
    best_ang = best_dball_ang;
  }

  if(smallest_error<2000)
  {
    //the best shot is close enough to th target, so use it
    pow=best_pow+0.5*normal();
    ang=best_ang;
    if(*shot==S_PONGER)
    {
      pow=max(pow,50.0);
      ang+=0.007*normal();
    }
  }
  else if ((m_y+3<=gp->getdepth()) && !random(3) && (gp->getpixel(m_x-1, m_y+3)!=W_RUBBER) &&
           (gp->getpixel(m_x, m_y+3)!=W_RUBBER) && (gp->getpixel(m_x+1, m_y+3)!=W_RUBBER))
  {
    //defend own position by laying down rubber
    *warhead = W_RUBBER;
    *shot = S_STANDARD;
    pow = 0;
    ang = 0;
  }
  else
  {
    //No decent shot found, so we tunnel.  Caluclate a suitable trajectory
    //(which probably has rocks, walls etc in the way), then blast through
    //it.  This may take a while, but it will eventually work.

    *warhead=W_HUGE;
    *shot = S_STANDARD;

    ang=atan(double(y_target)/abs(x_target));
    ang+=(M_PI/2-ang)/2.0;
    ang=max(ang,M_PI*0.33333);

    //Avoid pinpoint accuracy.
    ang+=0.007*normal();

    //x(t) = pow*cos(ang)*t
    //y(t) = pow*sin(ang)*t+g*t^2/2
    //Eliminate t by sustituting first equation into second.
    //y(t) = sin(ang)*x(t)/cos(ang)+g*x(t)^2/(pow^2*cos(ang)^2)
    //
    //pow=sqrt[g*x^2/(2*cos(ang)^2*(y-tan(ang)*x))]

    pow=-5.0*x_target*x_target/cos(ang)/cos(ang)/(-y_target-tan(ang)*abs(x_target));
    if(pow>0)pow=sqrt(pow);else pow=42;


    *dxp=pow*cos(ang)*(x_target>0?1:-1);
    *dyp=pow*sin(ang);
    trytp=*shot;
    int x_bounce, y_bounce;
    gp->fire(m_x,m_y,*dxp,-*dyp,trytp,&tryxf,&tryyf,false, &x_bounce, &y_bounce);

    //if the first bounce was near the target...
    if (x_bounce >=0 && x_bounce <=gp->getwidth() &&
        y_bounce >=0 && y_bounce <=gp->getdepth() &&
        (x_bounce-x_target-m_x)*(x_bounce-x_target-m_x)+(y_bounce-y_target-m_y)*(y_bounce-y_target-m_y)<2000)
    {
      //Store the firing data.  This is useful if we're firing at a rubbered opponent.  The shot
      //will bounce off, damaging one pixel of rubber.  Then next time, we'll use exactly the
      //same power and angle, so we'll hit the damaged pixel.
      m_firing_data = new firing_data;
      m_firing_data->dx=*dxp;
      m_firing_data->dy=*dyp;
      m_firing_data->xstart=m_x;
      m_firing_data->ystart=m_y;
      m_firing_data->xtarget=x_target;
      m_firing_data->ytarget=y_target;
      m_firing_data->xend=x_bounce;
      m_firing_data->yend=y_bounce;
    }

    if(abs(tryxf-m_x)<30)
      if(random(10))
        *warhead=W_DIRTBUSTER;
      else
        //This is an attempt to make the AI perform the "Stewart Maneouvre":
        //bouncing rubber off the enemy's rubber defence (damaging it in the process)
        //and rubbering its own position.
        *warhead=W_RUBBER;
    else if(tryxf>0&&tryxf<gp->getwidth()&&tryyf>50&&tryyf<gp->getdepth()&&
            !gp->getpixel(tryxf+(x_target>0?1:-1),tryyf)&&
            gp->getpixel(tryxf,tryyf-50)&&abs(tryxf-m_x-x_target)>30)
    {
      //use the W_HOLE to shoot through walls
      *warhead=W_HOLE;
    }
    else
      *warhead=W_HUGE;
  }

  //Don't use special warhead if we've ran out!
  if (!m_nspecials && is_special_warhead(*warhead))
    {
    if (*warhead==W_RUBBER) *warhead=W_SINKWALL; else *warhead=W_MEDIUM;
    }

  //Sanity
  if (*shot == S_SINKER || *shot == S_FLOATER || *shot==S_GROUNDBURST)
      *warhead = W_HUGE;

  *dxp=pow*cos(ang)*(x_target>0?1:-1);
  *dyp=pow*sin(ang);
}


//Try to use old firing data to complete a 2-shot tactic.  Return true if a shot is selected.
bool player::computer_use_old_firing_data(double *dxp, double *dyp, int *shot, int *warhead, players *pp, ground *gp)
{
  struct delete_it //and zero it!
  {
    delete_it(firing_data *&p):m_p(p){}
    firing_data *&m_p;
    ~delete_it(){delete m_p; m_p=0;}
  };
  delete_it di(m_firing_data);

  if (!m_firing_data) return false;
  if (m_x != m_firing_data->xstart || m_y != m_firing_data->ystart) return false;
  bool found = false;
  players::iterator i = pp->first_player();
  do
  {
    int x_target, y_target;
    (*i)->getcoords(&x_target, &y_target);
    x_target-=m_x;
    y_target-=(m_y+1);
    if (x_target==m_firing_data->xtarget && y_target==m_firing_data->ytarget)
    {
      found = true;
      break;
    }
    i = pp->next_player(i);
  }while(i!=pp->first_player());

  if (!found) return false;

  int x = m_firing_data->xend;
  int y = m_firing_data->yend;
  if (x<0 || x>gp->getwidth() || y<0 || y>gp->getdepth()) return false;

  if (is_dirt(gp->getpixel(x,y)) || gp->getpixel(x,y)==T_GREEN_STUFF ||
      gp->getpixel(x,y)==T_TANK_COLOUR)
  {
    *dxp = m_firing_data->dx;
    *dyp = m_firing_data->dy;
    *shot = S_STANDARD;
    *warhead = W_HUGE; //todo: be smarter about this one
    return true;
  }

  return false;
}

void player::human_fire(double *dxp,double *dyp,int *shot,int *warhead,
                           players *pp, ground *gp, bool *stubborn)
{
  bool stupid;
  do
  {
    {
      boringbox bb(m_defaults.left,m_defaults.top,m_defaults.left+240,
                   m_defaults.top+110,64,("It's "+m_name+"'s turn.").c_str());

      bb.spawn(new text(m_defaults.left+10,m_defaults.top+20,64,
               (string("Score: ")+int2str(m_score)+" Armour: "+int2str(m_armour)).c_str()));

      bb.spawn(new text(m_defaults.left+10,m_defaults.top+30,64,
               (string("Specials: ")+int2str(m_nspecials)).c_str()));
      bb.spawn(new text(m_defaults.left+10,m_defaults.top+48,64,"Power:"));
      textbox *powbox;
      bb.spawn(powbox=new textbox(m_defaults.left+64,m_defaults.top+46,5,64,65,
                                  m_defaults.power));
      bb.spawn(new text(m_defaults.left+10,m_defaults.top+63,64,"Angle:"));
      textbox *angbox;
      bb.spawn(angbox=new textbox(m_defaults.left+64,m_defaults.top+61,5,64,65,
                                  m_defaults.angle));
      menu *lrmenu;
      char *lorr[]={"Left", "Right"};
      bb.spawn(lrmenu=new menu(m_defaults.left+108,m_defaults.top+61,64,2,
                                  m_defaults.direction,lorr));

      bb.spawn(new text(m_defaults.left+10,m_defaults.top+78,64,"Shot:"));
      menu *smenu;
      bb.spawn(smenu=new menu(m_defaults.left+108, m_defaults.top+76, 64,
                              m_nspecials?sizeof(shots)/sizeof(*shots):non_special_shots,
                              m_defaults.shot, shots));

      bb.spawn(new text(m_defaults.left+10,m_defaults.top+93,64,"Warhead:"));
      menu *wmenu;
      bb.spawn(wmenu=new menu(m_defaults.left+108, m_defaults.top+91, 64,
                              m_nspecials?sizeof(warheads)/sizeof(*warheads):non_special_warheads,
                              m_defaults.warhead, warheads));

      bb.spawn(new button(m_defaults.left+170,m_defaults.top+40,m_defaults.left+220,
                          m_defaults.top+70,68,"FIRE!",stoppit1));
      bb.redraw();
      void (*oldfun)()=uiobject::uol.catchchar[13];
      uiobject::uol.catchchar[13]=uiobject::uol.donothing;
      uiobject::uol.wait();
      uiobject::uol.catchchar[13]=oldfun;
      strcpy(m_defaults.power,powbox->getstr());
      strcpy(m_defaults.angle,angbox->getstr());
      m_defaults.direction=lrmenu->getcurrent();
      m_defaults.left=bb.getl();
      m_defaults.top=bb.gett();
      m_defaults.shot = smenu->getcurrent();
      m_defaults.warhead = wmenu->getcurrent();
      double pow=atof(powbox->getstr());
      if(smenu->getcurrent()==S_PONGER)pow=max(pow,40.0);
      pow=min(max(0.0,pow),1000.0);
      double ang=atof(angbox->getstr())*M_PI/180;
      ang=max(-1000.0,min(ang,1000.0));
      *dxp=pow*cos(ang)*(2*lrmenu->getcurrent()-1);
      *dyp=pow*sin(ang);
      *shot=smenu->getcurrent();
      *warhead=wmenu->getcurrent();

      //check for stupidity
      stupid = (pow>200);
      stupid |= ang>M_PI*0.45;//why shoot backwards?
      int xmin, xmax;
      pp->getlimits(&xmin,&xmax);
      stupid |= (lrmenu->getcurrent()&&m_x==xmax||!lrmenu->getcurrent()&&m_x==xmin);

      bool shoot_self = (*shot==S_STANDARD||*shot==S_PONGER||*shot==S_AIRBURST) &&
        (is_dirt(gp->getpixel(m_x,m_y)) || gp->getpixel(m_x,m_y)==T_GREEN_STUFF||gp->getpixel(m_x,m_y)==T_RUBBER);
      shoot_self |=(*shot==S_SINKER||*shot==S_FLOATER)&&
            (gp->getpixel(m_x,m_y)==T_GREEN_STUFF||gp->getpixel(m_x,m_y)==T_RUBBER);
      shoot_self |= (pow<10);
      //It's not "stupid" to hit yourself with a dirtbuster or rubber.  Perhaps W_DIRTBALL
      //should be added to this list.
      stupid |= (shoot_self && (*warhead!=W_DIRTBUSTER && *warhead!=W_RUBBER));
    }

    if(stupid)
    {
      boringbox bb(60,100,260,160,64,"Stupidity Alert");
      bb.spawn(new text(110,120,64,"Are you sure?"));
      bb.spawn(new button(70,140,100,150,64,"Yes",stoppit1));
      bb.spawn(new button(220,140,250,150,64,"No",stoppit2));
      bb.redraw();
      uiobject::uol.wait();
      if(*stubborn=(uiobject::uol.stopwaiting==1))stupid=0;
    }
    else *stubborn=0;
  }
  while(stupid);
}


void player::down1()
{
  m_y++;
  m_fall_damage++;
}

void player::getcoords(int *xp,int *yp){*xp=m_x; *yp=m_y;}
void player::blast(int b){m_blast_damage=b;}

//Display the damage report, apply the actual damage.
void player::dr(player *inflictor,bool stubborn)
{

  int h,sc;
  h=sc=0;

  if(m_blast_damage||m_fall_damage)
  {
    boringbox bb(40,25,280,175,64,("Damage report for "+m_name).c_str());
    bb.spawn(new text(50,55,64,"Original armour:"));
    string armour = int2str(m_armour);
    bb.spawn(new text(264-8*armour.size(),55,64,armour.c_str()));
    if(m_blast_damage)
    {
      h++;
      bb.spawn(new text(50,55+10*h,64,"Blast damage:"));
      string bd = int2str(-m_blast_damage);
      bb.spawn(new text(264-8*bd.size(),55+10*h,64,bd.c_str()));
      m_armour-=m_blast_damage;
      sc+=m_blast_damage;
      m_blast_damage=0;
    }
    if(m_fall_damage)
    {
      h++;
      bb.spawn(new text(50,55+10*h,64,"Fall damage:"));
      string fd = int2str(-m_fall_damage);
      bb.spawn(new text(264-8*fd.size(),55+10*h,64,fd.c_str()));
      m_armour-=m_fall_damage;
      sc+=m_fall_damage;
      m_fall_damage=0;
      if(m_rock_bottom)
      {
        h++;
        bb.spawn(new text(50,55+10*h,64,"Rock bottom penalty:"));
        bb.spawn(new text(240,55+10*h,64,"-20"));
        m_armour-=20;
        sc+=20;
        m_rock_bottom=false;
      }
    }
    if(inflictor==this)
    {
      h++;
      bb.spawn(new text(50,55+10*h,64,"Suicide penalty:"));
      bb.spawn(new text(240,55+10*h,64,"-20"));
      m_armour-=20;
      sc=0;
      if(stubborn)
      {
        h++;
        bb.spawn(new text(50,55+10*h,64,"I told you so:"));
        bb.spawn(new text(240,55+10*h,64,"-10"));
        m_armour-=10;
      }
    }
    h++;
    bb.spawn(new text(50,55+10*h,64,"Remaining armour:"));
    bb.spawn(new text(264-8*int2str(m_armour).size(),55+10*h,64,int2str(m_armour).c_str()));
    if(m_armour<0)
    {
      h++;
      inflictor->m_kills++;
      if(inflictor==this) m_has_killed_self=true; //whoops!
      bb.spawn(new text(50,55+10*h,64,(m_name+" is dead").c_str()));
    }
    bb.spawn(new button(250,155,270,165,64,"OK",stoppit1));
    bb.redraw();
    #ifndef SPEEEED
    uiobject::uol.wait();
    #else
    rest(500);
    #endif
  }
  inflictor->m_score+=sc;
}


void player::rock_bottom_check(int h) {m_rock_bottom=(m_y==h);}
bool player::is_dead(){return m_armour<0;}
const string &player::get_name(){return m_name;}
int player::get_score(){return m_score;}
void player::inc_score(int i){m_score+=i;}
int player::get_kills(){return m_kills;}
bool player::has_killed_self(){return m_has_killed_self;}


players::players(int n,ground &g)
:m_ground(g)
{
  for(int i=0; i<n; ++i)
    m_players.insert(m_players.end(), new player(n,i+1,&g));
  m_current = m_players.begin();
}

players::~players()
{
  for (std::list<player *>::iterator i=m_players.begin(); i!=m_players.end(); ++i)
    delete *i;
  for (std::deque<player *>::iterator i=m_morgue.begin(); i!=m_morgue.end(); ++i)
    delete *i;
}

bool less_by_score(player *_1, player *_2)
{return _1->get_score() < _2->get_score();}

void players::go()
{
  double dx,dy;
  int xend,yend;
  bool stubborn;

  do
  {
    int x,y;
    (*m_current)->getcoords(&x,&y);
    m_ground.scrollto(x,y);
    int shot, warhead;
    (*m_current)->fire(&dx,&dy,&shot,&warhead,this,&m_ground,&stubborn);
    if(shot==S_PONGER)
      sound::play(PONG1,255,128,1000,0);
    else
      sound::play(SHOOT,min(255,(int)(3*sqrt(dx*dx+dy*dy))),128,1000,0);

    std::list<player *>::iterator next=next_player(m_current);
    player *shooter = *m_current;
    if(m_ground.fire(x,y,dx,-dy,shot,&xend,&yend,1))
    {
      m_ground.detonate(this,xend,yend,warhead);
      //Display damage reports, remove dead players from m_players, put them in m_morgue.
      for(std::list<player *>::iterator i=m_players.begin(); i!=m_players.end();/* no i++! */)
      {
        (*i)->dr(shooter,stubborn);
        if((*i)->is_dead())
        {
          m_ground.drawoff(*i); //draw the corpse
          if(next==i) next=next_player(next); //Make sure "next" points to a player who isn't dead.
          m_morgue.insert(m_morgue.end(), *i);
          m_players.erase(i++);
          int x,y;
          (*m_morgue.rbegin())->getcoords(&x,&y);
          m_ground.landfall(x-2, x+2, this);
        }
        else ++i;
      }
    }
    setsky();
    m_current = next;
  }while(get_nplayers()>1);

  boringbox bb(40,50,280,150,64,"Game over!");

  if(get_nplayers()==1)
  {
    //We have a winner!  Since m_current points to a valid player, it must be the winner.
    (*m_current)->inc_score(500);
    bb.spawn(new text(60,70,64,((*m_current)->get_name()+" is the winner.").c_str()));
    //As a reward, the winner is ceremonially killed and placed on the morgue.
    m_morgue.insert(m_morgue.end(),*m_current);
    m_players.erase(m_current);
    m_current = m_players.end();
  }
  else
  { 
    bb.spawn(new text(60,70,64,"You're all dead."));
  }

  bb.spawn(new button(250,128,270,140,64,"OK",stoppit1));
  bb.redraw();
  uiobject::uol.wait();
  fade_out(1);

  std::sort(m_morgue.rbegin(), m_morgue.rend(), less_by_score);

  clear(screen);
  setcol(64,63,63,63);

  int t=0;
  for(std::deque<player *>::iterator i=m_morgue.begin(); i!=m_morgue.end(); ++i)
  {
    if(!(t%16))
    {
      clear(screen);
      textout(screen,font,"Player    Score    Kills",10,5,64);
    }


    textout(screen,font,(*i)->get_name().c_str(),10,t%16*10+30,64);

    string score=int2str((*i)->get_score());
    textout(screen,font,score.c_str(),130-8*score.size(),t%16*10+30,64);

    string kills =int2str((*i)->get_kills());
    textout(screen,font,kills.c_str(),178-8*kills.size(),t%16*10+30,64);

    if((*i)->has_killed_self())
      textout(screen,font,"(Including self)",186,t%16*10+30,64);
    t++;


    if(!(t%16 && t!=int(m_morgue.size())))
    {
      textout(screen,font,"Press any key",100,191,64);
      while(keypressed()) readkey();
      readkey();
      clear(screen);
    }
  }
}

int players::fall()
{
int b=1;
for(std::list<player *>::iterator i=m_players.begin(); i!=m_players.end(); ++i)
  {
  b*=!m_ground.fallplayer(*i);
  }
return !b;
}

int players::get_nplayers(){return m_players.size();}

void players::getlimits(int *xmin,int *xmax)
{
  int x,y;
  for (std::list<player *>::iterator i=m_players.begin(); i!=m_players.end(); ++i)
  {
    (*i)->getcoords(&x,&y);
    *xmin=min(*xmin,x);
    *xmax=max(*xmax,x); 
  }
}


std::list<player *>::iterator players::first_player()
{
  return m_players.begin();
}


//return (an iterator to) the player whose turn it is next.
std::list<player *>::iterator players::next_player(std::list<player *>::iterator i)
{
  //?
  if (i==m_players.end()) return m_players.end();

  ++i;
  if (i==m_players.end()) return m_players.begin();
  return i;
}



void players::rbcheck(int h)
{
  for (std::list<player *>::iterator i=m_players.begin(); i!=m_players.end(); ++i)
  {
    (*i)->rock_bottom_check(h);
  }
}

void players::redrawall()
{
  for (std::list<player *>::iterator i=m_players.begin(); i!=m_players.end(); ++i)
  {
    m_ground.drawplayer(*i);
  }
}

void players::getvictim(player *not_this_one,int *xp,int *yp)
{
  int n=random(m_players.size()-1);
  iterator p=first_player();
  if(*p==not_this_one) p=next_player(p);
  for(int i=0; i<n; ++i)
  {
    p=next_player(p);
    if(*p==not_this_one) p=next_player(p);
  }
  (*p)->getcoords(xp,yp);
}


int main()
{
  allegro_init();
  install_timer();
  if(install_mouse()==-1)
  {
    printf("you don't have a mouse driver!\n");
    return 1;
  };
  set_gfx_mode(GFX_AUTODETECT,320,240,0,0);
  install_keyboard();
  uiobject::uol.mvs();
  srand(time(NULL));


  while(1)
  {
    int nplayers;
    bool sound;
    do
    {
      clear(screen);
      for(int a=0;a<64;a++)
      {
        if(a)setcol(a,63-a,3*(63-a)/4+12,2*(63-a)/3);
        setcol(a+128,63,a,a/8);
      }
      setcol(64,63,63,63);
      setcol(65,31,31,31);
      setcol(66, 0,63, 0);
      setcol(68,40, 0, 0);
      setcol(69, 0,31,31);
      setsky();
      textout(screen,font,"Players:",10,50,64);
      textbox numbox(82,48,2,64,65,"3");
      numbox.activate();
      button okbutton(200,60,220,70,64,"OK",stoppit1);
      okbutton.redraw();
      button quitbutton(200,90,238,100,64,"QUIT",stoppit2);
      quitbutton.redraw();
      textout(screen,font,"Sound:",10,100,64);
      char *sm[2]={"Perhaps","No"};
      menu soundmenu(82,98,64,2,0,sm);
      soundmenu.redraw();
      uiobject::uol.wait();
      if(uiobject::uol.stopwaiting==2)//Quit button pressed
      {
        goto the_end;
      }
      nplayers=atoi(numbox.getstr());
      sound = !soundmenu.getcurrent();
      clear(screen);
    }while(nplayers<2||nplayers>99);

    if(sound)
    {
      switch(sound::init())
      {
        //case 0:/*no problems*/break;
        case 1:
          boringbox bb(50,50,270,100,15,"Sound error");
          bb.spawn(new text(60,65,15,"Could not install"));
          bb.spawn(new button(245,85,265,95,15,"OK",stoppit1));
          bb.redraw();
          uiobject::uol.wait();
          break;
      }
    }
    else sound::init_off();

    try
    {
      ground the_ground(1200,400,10,10,SCREEN_W-11,SCREEN_H-11);
      uiobject::uol.catchechar[KEY_X]=throw_alt_X;
      players the_players(nplayers, the_ground);
      uiobject::uol.catchchar[13]=stoppit1wrapper;

      the_players.go();

      uiobject::uol.catchchar[13]=&uiobjectlist::donothing;
      uiobject::uol.catchechar[45]=&uiobjectlist::donothing;
    }
    catch(alt_X)
    {
      set_gfx_mode(GFX_TEXT,80,25,0,0);
      printf("Thank you for pressing the self-destruct button.\n\n");
      goto the_end;
    }
  }
  
  the_end:
  sound::finish();
  uiobject::uol.kvs();
  return 0;
}
END_OF_MAIN()

void stoppit1(button *){uiobject::uol.stopwaiting=1;}
void stoppit2(button *){uiobject::uol.stopwaiting=2;}

//N(0,1) distribution.
double normal()
{
  //box-muller method.
  int r;
  double d=RAND_MAX;
  while(!(r=rand()));
  return cos(2.0*M_PI*rand()/d) * sqrt(2.0*log(d/r));
}

void stoppit1wrapper(){stoppit1(0);}

//This is really cute: sky colour depends on local time.
void setsky()
{
  time_t timer;
  tm *t;
  byte n,r,g,b;
  timer=time(NULL);
  t=localtime(&timer);
  n=(byte)(31.0-31.0*cos(((double)(t->tm_min)/60.0+t->tm_hour)/24.0*2.0*M_PI));
  r=g=n*n/80;b=n;
  for(n=0;n<64;n++)setcol(192+n,r*n/64,g*n/64,b);
}

void setcol(byte n,byte r,byte g,byte b)
{
  RGB col;
  col.r=r;col.g=g;col.b=b;
  set_color(n,&col);
}

void throw_alt_X() {throw alt_X();}
