/*
    Copyright (C) 2004 Mark Boyd

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be fun to play,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef TANX_H
#define TANX_H

#include <deque>
#include <list>

typedef unsigned char byte;

class player;
class players;


class ground
{
public:
  ground(int width, int depth, int left,int top, int right, int bottom);
  ~ground();
  void draw();
  void moveto(int,int);
  void scrollto(int,int);
  void addplayer(int n,int nplayers,int *xp,int *yp);
  void drawplayer(player *);
  void drawoff(player *);
  bool fallplayer(player *);
  bool fire(int x_start, int y_start, double x_speed, double y_speed,
            int shot, int *x_end, int *y_end, int is_real=true, int *x_bounce=0, int *y_bounce=0);
  void detonate(players *, int x, int y, int warhead);
  void circlebang(int,int,int,int,players &);
  int getdepth();
  int getwidth();
  void landfall(int,int,players *);
  byte getpixel(int,int);

private:
  int m_width,m_depth;
  int m_left,m_top,m_right,m_bottom; 
  BITMAP *m_bitmap,*m_buffer;
  int m_xdisp,m_ydisp; 
  byte *m_sky_table;

  void annulus(int,int,int,int,int);
};



class player
{
public:
  player(int nplayers, int n,ground *);
  ~player();
  void fire(double *,double *,int *, int *, players *,ground *,bool *);
  void down1();
  void getcoords(int *,int *);
  void blast(int);
  void dr(player *,bool);
  void rock_bottom_check(int h);
  bool is_dead();
  const string &get_name();
  int get_score();
  void inc_score(int);
  int get_kills();
  bool has_killed_self();

private:
  struct defs
  {
    char power[6];
    char angle[6];
    short direction;
    short shot;
    short warhead;
    short left;
    short top;
  };
  enum type {HUMAN, COMPUTER};
  struct firing_data
  {
    double dx;
    double dy;

    int xstart;
    int ystart;

    int xtarget;
    int ytarget;

    int xend;
    int yend;
  };

  void computer_fire(double *dxp,double *dyp,int *shot,int *warhead,
                           players *pp, ground *gp);
  void computer_choose_shot(double *dxp, double *dyp, int *shot,
                            int *warhead, players *pp, ground *gp);
  bool computer_use_old_firing_data(double *dxp, double *dyp, int *shot,
                                 int *warhead, players *pp, ground *gp);
  void human_fire(double *dxp,double *dyp,int *shot,int *warhead,
                           players *pp, ground *gp, bool *stubborn);

  std::string m_name;
  int m_x,m_y;
  int m_armour;
  int m_score;
  int m_kills;
  bool m_has_killed_self;
  int m_blast_damage, m_fall_damage;
  bool m_rock_bottom;
  type m_type;
  int m_nspecials;
  defs m_defaults;
  firing_data *m_firing_data;
};




class players
{
public:
  typedef std::list<player *>::iterator iterator;

  players(int,ground &);
  ~players();
  void go();
  int fall();
  int get_nplayers(void);
  void getlimits(int *,int *);
  void rbcheck(int);
  void redrawall();
  void getvictim(player *,int *,int *);
  iterator next_player(iterator);
  iterator first_player();

private:
  ground &m_ground;
  std::list<player *> m_players;
  std::list<player *>::iterator m_current;
  std::deque<player *> m_morgue;

};


class sound
{
public:
  static int d,m;
  static int init();
  static void finish();
  static int play(int,int,int,int,int);
  static void sound::stop(int);
  static void sound::init_off(void);
private:
  static DATAFILE *df;
  static int ok;
};


class alt_X{};//exception


void setcol(byte,byte,byte,byte);
void stoppit1(button *);
void stoppit2(button *);
void stoppit1wrapper(void);
double normal(void);
void setsky(void);
void throw_alt_X(void);


#endif
